import unittest
import os
import subprocess as sp
import time

from reloading import reloading

SRC_FILE_NAME = 'temporary_testing_file.py'

TEST_CHANGING_SOURCE_LOOP_CONTENT = '''
from reloading import reloading
from time import sleep

for epoch in reloading(range(10)):
    sleep(0.1)
    print('INITIAL_FILE_CONTENTS')
'''

TEST_CHANGING_SOURCE_FN_CONTENT = '''
from reloading import reloading
from time import sleep

@reloading
def reload_this_fn():
    print('INITIAL_FILE_CONTENTS')

for epoch in reloading(range(10)):
    sleep(0.1)
    reload_this_fn()
'''

TEST_KEEP_LOCAL_VARIABLES_CONTENT = '''
from reloading import reloading
from time import sleep

fpath = "DON'T CHANGE ME"
for epoch in reloading(range(1)):
    assert fpath == "DON'T CHANGE ME"
'''

TEST_PERSIST_AFTER_LOOP = '''
from reloading import reloading
from time import sleep

state = 'INIT'
for epoch in reloading(range(1)):
    state = 'CHANGED'

assert state == 'CHANGED'
'''

TEST_COMMENT_AFTER_LOOP_CONTENT = '''
from reloading import reloading
from time import sleep

for epoch in reloading(range(10)):
    sleep(0.1)
    print('INITIAL_FILE_CONTENTS')

# a comment here should not cause an error
'''

TEST_FORMAT_STR_IN_LOOP_CONTENT = '''
from reloading import reloading
from time import sleep

for epoch in reloading(range(10)):
    sleep(0.1)
    file_contents = 'FILE_CONTENTS'
    print(f'INITIAL_{file_contents}')
'''


def run_and_update_source(init_src, updated_src=None, update_after=0.5):
    '''Runs init_src in a subprocess and updates source to updated_src after
    update_after seconds. Returns the standard output of the subprocess and
    whether the subprocess produced an uncaught exception.
    '''
    with open(SRC_FILE_NAME, 'w') as f:
        f.write(init_src)

    cmd = ['python3', SRC_FILE_NAME]
    with sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE) as proc:
        if updated_src != None:
            time.sleep(update_after)
            with open(SRC_FILE_NAME, 'w') as f:
                f.write(updated_src)

        try:
            stdout, _ = proc.communicate(timeout=2)
            stdout = stdout.decode('utf-8')
            has_error = False
        except:
            stdout = ''
            has_error = True
            proc.terminate() 

    if os.path.isfile(SRC_FILE_NAME):
        os.remove(SRC_FILE_NAME)

    return stdout, has_error


class TestReloading(unittest.TestCase):

    def test_simple_looping(self):
        iters = 0
        for _ in reloading(range(10)):
            iters += 1

    def test_changing_source_loop(self):
        stdout, _ = run_and_update_source(
          init_src=TEST_CHANGING_SOURCE_LOOP_CONTENT,
          updated_src=TEST_CHANGING_SOURCE_LOOP_CONTENT.replace('INITIAL', 'CHANGED').rstrip('\n'))

        self.assertTrue('INITIAL_FILE_CONTENTS' in stdout and
                        'CHANGED_FILE_CONTENTS' in stdout)

    def test_comment_after_loop(self):
        stdout, _ = run_and_update_source(
          init_src=TEST_COMMENT_AFTER_LOOP_CONTENT,
          updated_src=TEST_COMMENT_AFTER_LOOP_CONTENT.replace('INITIAL', 'CHANGED').rstrip('\n'))

        self.assertTrue('INITIAL_FILE_CONTENTS' in stdout and
                        'CHANGED_FILE_CONTENTS' in stdout)

    def test_format_str_in_loop(self):
        stdout, _ = run_and_update_source(
          init_src=TEST_FORMAT_STR_IN_LOOP_CONTENT,
          updated_src=TEST_FORMAT_STR_IN_LOOP_CONTENT.replace('INITIAL', 'CHANGED').rstrip('\n'))

        self.assertTrue('INITIAL_FILE_CONTENTS' in stdout and
                        'CHANGED_FILE_CONTENTS' in stdout)

    def test_keep_local_variables(self):
        _, has_error = run_and_update_source(init_src=TEST_KEEP_LOCAL_VARIABLES_CONTENT)
        self.assertFalse(has_error)

    def test_persist_after_loop(self):
        _, has_error = run_and_update_source(init_src=TEST_PERSIST_AFTER_LOOP)
        self.assertFalse(has_error)

    def test_simple_function(self):
        @reloading
        def some_func():
            return 'result'
        
        self.assertTrue(some_func() == 'result')

    def test_changing_source_function(self):
        stdout, _ = run_and_update_source(
          init_src=TEST_CHANGING_SOURCE_FN_CONTENT,
          updated_src=TEST_CHANGING_SOURCE_FN_CONTENT.replace('INITIAL', 'CHANGED').rstrip('\n'))

        self.assertTrue('INITIAL_FILE_CONTENTS' in stdout and
                        'CHANGED_FILE_CONTENTS' in stdout)


if __name__ == '__main__':
    unittest.main()



import time
import re
import inspect
import sys
import ast
import traceback
import types
from itertools import chain

def reloading(fn_or_seq):
    '''Wraps a loop iterator or decorates a function to reload source code.

    A function that when wrapped around the outermost iterator in a for loop,
    causes the loop body to reload from source before every iteration while
    keeping the state.
    When used as a function decorator, the function is reloaded from source 
    before each execution.

    Args:
        fn_or_seq (function | iterable): A function or loop iterator which should
            be reloaded from source before each execution or iteration,
            respectively
    '''
    if isinstance(fn_or_seq, types.FunctionType):
        return _reloading_function(fn_or_seq)
    return _reloading_loop(fn_or_seq)


def find_loop(tree):
    for child in ast.walk(tree):
        if not isinstance(child, ast.For):
            continue
        if not isinstance(child.iter, ast.Call):
            continue
        if child.iter.func.id == 'reloading':
            return child


def locate_loop_body(module, loop):
    ends = set([ node.lineno 
                    for node in ast.walk(module) 
                    if hasattr(node, 'lineno') and node.lineno > loop.lineno ])

    starts = set()

    def visit(node):
        if hasattr(node, 'lineno') and node.lineno > loop.lineno:
            starts.add(node.lineno)
            if node.lineno in ends:
                ends.remove(node.lineno)
        for child in ast.iter_child_nodes(node):
            visit(child)

    for stmt in loop.body:
        visit(stmt)

    if len(ends) == 0: 
        return min(starts), -1

    return min(starts), min(ends)


def unique_name(used):
    return max(used, key=len) + "0"


def _reloading_loop(seq):
    frame = inspect.currentframe()

    caller_globals = frame.f_back.f_back.f_globals
    caller_locals = frame.f_back.f_back.f_locals
    unique = unique_name(chain(caller_locals.keys(), caller_globals.keys()))
    for j in seq:
        fpath = inspect.stack()[2][1]
        with open(fpath, 'r') as f:
            src = f.read() + '\n'

        # find the iteration variables in the caller module's source
        match = re.search('\s*for (.+?) in reloading', src)
        if match is None:
            break 
        itervars = match.group(1)

        # find the loop body in the caller module's source
        tree = ast.parse(src)
        loop = find_loop(tree)
        start, end = locate_loop_body(tree, loop)
        lines  = src.split('\n')
        if end < 0:
            end = len(lines)
        body_lines = lines[start-1:end-1] # -1 as line numbers are 1-indexed

        # remove indent from lines in loop body, only if a line
        # starts with this indentation as comments might not
        indent = re.search('([ \t]*)\S', body_lines[0])
        body = '\n'.join([ line[len(indent.group(1)):] 
            for line in body_lines 
            if line.startswith(indent.group(1))])

        caller_locals[unique] = j
        exec(itervars + ' = ' + unique, caller_globals, caller_locals)

        try:
            # run main loop body
            exec(body, caller_globals, caller_locals)
        except Exception:
            exc = traceback.format_exc()
            exc = exc.replace('File "<string>"', 'File "{}"'.format(fpath))
            sys.stderr.write(exc + '\n')
            print('Edit {} and press return to continue with the next iteration'.format(fpath))
            sys.stdin.readline()

    return []


def find_function_in_source(fn_name, src):
    '''Finds line number of start and end of a function with a 
    given name within the given source code.
    '''
    tree = ast.parse(src)

    # find the parent of the function definition so that we can find out
    # where the function definition ends by using the starting line
    # number of the subsequent child after the function definition
    for parent in ast.walk(tree):
        fn_end = len(src.split('\n'))

        for child in reversed(list(ast.iter_child_nodes(parent))):
            if not isinstance(child, ast.FunctionDef)\
               or child.name != fn_name\
               or not hasattr(child, 'decorator_list')\
               or len([ 
                   dec 
                   for dec in child.decorator_list 
                   if dec.id == 'reloading' ]) < 1:

                if hasattr(child, 'lineno'):
                    fn_end = child.lineno - 1
                continue

            # if we arrived here, child is the function definition
            fn_start = child.lineno
            return fn_start, fn_end, child.col_offset

    return -1, -1, 0


def _reloading_function(fn):
    frame, fpath = inspect.stack()[2][:2]
    caller_locals = frame.f_locals
    caller_globals = frame.f_globals

    # if we are redefining the function, we need to load the file path
    # from the function's dictionary as it would be `<string>` otherwise
    # which happens when defining functions using `exec`
    if fn.__name__ in caller_locals:
        fpath = caller_locals[fn.__name__].__dict__['__fpath__']

    def wrapped(*args, **kwargs):
        with open(fpath, 'r') as f:
            src = f.read()

        start, end, indent = find_function_in_source(fn.__name__, src)
        lines = src.split('\n')
        fn_src = '\n'.join([ l[indent:] for l in lines[start-1:end] ])

        while True:
            try:
                exec(fn_src, caller_globals, caller_locals)
                break
            except Exception:
                exc = traceback.format_exc()
                exc = exc.replace('File "<string>"', 'File "{}"'.format(fpath))
                sys.stderr.write(exc + '\n')
                print('Edit {} and press return to try again'.format(fpath))
                sys.stdin.readline()


        # the newly defined function will also be decorated 
        # with `reloading` and, hence, we call the inner function without
        # triggering another reload (and another one...)
        inner = caller_locals[fn.__name__].__dict__['__inner__']
        return inner(*args, **kwargs)

    # save the inner function to be able to call it without 
    # triggering infinitely recursive reloading
    wrapped.__dict__['__inner__'] = fn
    # save the file path for later, as the original file path gets
    # lost by reloading and redefining the function using `exec`
    wrapped.__dict__['__fpath__'] = fpath
    wrapped.__name__ = fn.__name__
    wrapped.__doc__ = fn.__doc__

    return wrapped



from distutils.core import setup

setup(
  name = 'reloading',         
  packages = ['reloading'],   
  version = '1.0.0',      
  license='MIT',        
  description = 'Reloads a loop body from source on each iteration without loosing state',
  author = 'Julian Vossen',
  author_email = 'pypi@julianvossen.de',
  url = 'https://github.com/julvo/reloading',
  download_url = 'https://github.com/julvo/reloading/archive/v1.0.0.tar.gz',
  keywords = ['reload', 'reloading', 'refresh', 'loop'],
  install_requires=[],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Utilities',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
)
